import logging
from re import compile

from mkdocs.config.base import ValidationError
from mkdocs.config.config_options import Type

RE_LOCALE = compile(r"(^[a-z]{2}_[A-Z]{2}$)|(^[a-z]{2}$)")

log = logging.getLogger("mkdocs.plugins." + __name__)


class Locale(Type):
    """
    Locale Config Option

    Validate the locale config option against a given Python type.
    """

    def _validate_locale(self, value):
        # we allow the special case of the default language
        if value == "default":
            return
        if not RE_LOCALE.match(value):
            raise ValidationError(
                "Language code values must be either ISO-639-1 lower case "
                "or represented with they territory/region/county codes, "
                f"received '{value}' expected forms examples: 'en' or 'en_US'."
            )

    def _get_lang_dict_value(self, lang_key, lang_value):
        """
        Normalize the 'languages' option dict with backward compatibility.

        We support legacy:

            languages:
              en: English
              fr: Français

        And the new way:

            languages:
              en:
                name: English
                build: true
              fr:
                name: Français
                build: true
        """
        allowed_keys = set(["name", "link", "build", "site_name", "fixed_link"])
        lang_config = {
            "build": True,
            "link": f"./{lang_key}/" if lang_key != "default" else "./",
            "fixed_link": None,
            "name": lang_key,
            "site_name": None,
        }
        if isinstance(lang_value, str):
            lang_config["name"] = lang_value
        elif isinstance(lang_value, dict):
            unsupported_keys = set(lang_value.keys()).difference(allowed_keys)
            if unsupported_keys:
                log.warning(
                    f"'plugins.i18n.languages.{lang_key}' unsupported options: {','.join(unsupported_keys)}"
                )
            for key in lang_config:
                if key in lang_value:
                    lang_config[key] = lang_value[key]
        return lang_config

    def run_validation(self, value):
        value = super().run_validation(value)
        # default_language
        if isinstance(value, str):
            self._validate_locale(value)
        # languages
        if isinstance(value, dict):
            languages = {}
            for lang_key, lang_value in value.items():
                self._validate_locale(lang_key)
                languages[lang_key] = self._get_lang_dict_value(lang_key, lang_value)
            value = languages
        return value



import logging
import os
from pathlib import Path
from re import compile
from urllib.parse import quote as urlquote

from mkdocs import utils
from mkdocs.structure.files import File, Files
from mkdocs.structure.nav import get_navigation

RE_LOCALE = compile(r"(^[a-z]{2}_[A-Z]{2}$)|(^[a-z]{2}$)")

log = logging.getLogger("mkdocs.plugins." + __name__)


class I18nFolderFiles(Files):
    """
    This class extends MkDocs' Files class to support links and assets that
    have a translated locale suffix.

    Since MkDocs relies on the file.src_path of pages and assets we have to
    derive the file.src_path and check for a possible .<locale>.<suffix> file
    to use instead of the link / asset referenced in the markdown source.
    """

    locale = None
    translated = False

    def append(self, file):
        """
        Since I18nFolderFile find their own language versions, we need to avoid adding
        them multiple times when a localized version of a file is considered.

        The first I18nFolderFile is sufficient to cover all their possible localized versions.
        """
        for inside_file in self:
            if inside_file.dest_path == file.dest_path:
                return
        super().append(file)

    def __contains__(self, path):
        """
        Return a bool stipulating whether or not we found a translated version
        of the given path or the path itself.

        Since our plugin automatically localize links, this is useful for the
        mkdocs.structure.pages / path_to_url() method to point to the localized
        version of the file, if present.
        """
        return True if self.get_file_from_path(path) else False

    def get_file_from_path(self, path):
        """Return a File instance with File.src_path equal to path."""
        expected_src_path = Path(path)
        root_folder = expected_src_path.parts[0]
        expected_src_paths = [
            expected_src_path,
            expected_src_path.relative_to(root_folder),
            Path(self.locale) / Path(expected_src_path),
        ]
        for src_path in filter(lambda s: Path(s) in expected_src_paths, self.src_paths):
            return self.src_paths.get(os.path.normpath(src_path))

    def get_localized_page_from_url(self, url, language):
        """Return the I18nFolderFile instance from our files that match the given url and language"""
        if language:
            url = f"{language}/{url}"
        url = url.rstrip(".") or "."
        for file in self:
            if not file.is_documentation_page():
                continue
            if file.url == url:
                return file


class I18nFolderFile(File):
    """
    This is a i18n aware version of a mkdocs.structure.files.File
    """

    def __init__(
        self,
        file_from,
        language,
        all_languages=None,
        default_language=None,
        docs_dir=None,
        site_dir=None,
        use_directory_urls=None,
    ) -> None:
        # preserved from mkdocs.structure.files.File
        # since they are not calculated
        self.abs_src_path = file_from.abs_src_path
        self.docs_dir = docs_dir
        self.page = file_from.page
        self.site_dir = site_dir
        self.src_path = file_from.src_path

        # i18n addons
        self.all_languages = all_languages
        self.alternates = {lang: None for lang in self.all_languages}
        self.default_language = default_language
        self.dest_language = language
        self.initial_abs_dest_path = file_from.abs_dest_path
        self.initial_abs_src_path = file_from.abs_src_path
        self.initial_dest_path = file_from.dest_path
        self.initial_src_path = file_from.src_path
        self.locale_suffix = None
        self.root_folder = Path(file_from.src_path).parts[0]

        # the name
        self.name = Path(self.initial_src_path).name
        self.dest_name = self.name

        if self.root_folder not in self.all_languages:
            # non localized root folder, file should be copied as-is
            # in the destination language path
            self.locale = self.dest_language
            self.dest_path = Path(self.locale) / Path(file_from.dest_path)
            self.abs_dest_path = Path(self.site_dir) / Path(self.dest_path)
        elif language == "":
            # default version file
            self.locale = self.default_language
            self.dest_path = Path(self.initial_dest_path).relative_to(self.locale)
            self.abs_dest_path = Path(self.site_dir) / Path(self.dest_path)
        else:
            # in localized folder file
            self.locale = self.dest_language
            self.dest_path = file_from.dest_path
            self.abs_dest_path = file_from.abs_dest_path

        # set url
        self.url = self._get_url(use_directory_urls)

        # set ourself as our own alternate
        self.alternates[self.dest_language or self.default_language] = self

        # mkdocs expects strings for those
        self.abs_dest_path = str(self.abs_dest_path)
        self.abs_src_path = str(self.abs_src_path)
        self.dest_path = str(self.dest_path)
        self.src_path = str(self.src_path)

    def __repr__(self):
        return (
            f"I18nFolderFile(src_path='{self.src_path}', abs_src_path='{self.abs_src_path}',"
            f" dest_path='{self.dest_path}', abs_dest_path='{self.abs_dest_path}',"
            f" name='{self.name}', locale_suffix='{self.locale_suffix}',"
            f" dest_language='{self.dest_language}', dest_name='{self.dest_name}',"
            f" url='{self.url}')"
        )

    @property
    def non_i18n_src_path(self):
        """
        Return the path of the given page without any suffix.
        """
        if self._is_localized() is None:
            non_i18n_src_path = Path(self.initial_src_path).with_suffix("")
        else:
            non_i18n_src_path = (
                Path(self.initial_src_path).with_suffix("").with_suffix("")
            )
        return non_i18n_src_path

    def _is_localized(self):
        """
        Returns the locale detected in the file's suffixes <name>.<locale>.<suffix>.
        """
        for language in self.all_languages:
            initial_file_suffixes = Path(self.initial_src_path).suffixes
            expected_suffixes = [f".{language}", Path(self.initial_src_path).suffix]
            if len(initial_file_suffixes) >= len(expected_suffixes):
                if (
                    # fmt: off
                    initial_file_suffixes[-len(expected_suffixes):]
                    == expected_suffixes
                ):
                    return language
        return None

    @property
    def suffix(self):
        return Path(self.initial_src_path).suffix

    def _get_name(self):
        """Return the name of the file without it's extension."""
        return (
            "index"
            if self.non_i18n_src_path.name in ("index", "README")
            else self.non_i18n_src_path.name
        )

    def _get_dest_path(self, use_directory_urls):
        """Return destination path based on source path."""
        parent, _ = os.path.split(self.src_path)
        if self.is_documentation_page():
            if use_directory_urls is False or self.name == "index":
                # index.md or README.md => index.html
                # foo.md => foo.html
                return os.path.join(parent, self.name + ".html")
            else:
                # foo.md => foo/index.html
                return os.path.join(parent, self.name, "index.html")
        else:
            return os.path.join(parent, self.dest_name)

    def _get_url(self, use_directory_urls):
        """Return url based in destination path."""
        url = str(self.dest_path).replace(os.path.sep, "/")
        dirname, filename = os.path.split(url)
        if use_directory_urls and filename == "index.html":
            if dirname == "":
                url = "."
            else:
                url = dirname + "/"
        if self.dest_language:
            if url == ".":
                url += "/"
            # else:
            #     url = "/" + url
        return urlquote(url)

    def url_relative_to(self, other):
        """Return url for file relative to other i18n file."""
        return utils.get_relative_url(
            self.url,
            other.url
            if (isinstance(other, File) or isinstance(other, I18nFolderFile))
            else other,
        )


def on_files(self, files, config):
    """"""
    main_files = I18nFolderFiles([])
    main_files.default_locale = self.default_language
    main_files.locale = self.default_language
    for language in self.all_languages:
        self.i18n_files[language] = I18nFolderFiles([])
        self.i18n_files[language].default_locale = self.default_language
        self.i18n_files[language].locale = language

    for fileobj in files:

        file_locale = Path(fileobj.src_path).parts[0]

        if file_locale not in self.all_languages:
            if config["docs_dir"] in fileobj.abs_src_path:
                i18n_ffile = I18nFolderFile(
                    fileobj,
                    "",
                    all_languages=self.all_languages,
                    default_language=self.default_language,
                    docs_dir=config["docs_dir"],
                    site_dir=config["site_dir"],
                    use_directory_urls=config.get("use_directory_urls"),
                )
                main_files.append(i18n_ffile)
                for language in self.all_languages:
                    i18n_ffile = I18nFolderFile(
                        fileobj,
                        language,
                        all_languages=self.all_languages,
                        default_language=self.default_language,
                        docs_dir=config["docs_dir"],
                        site_dir=config["site_dir"],
                        use_directory_urls=config.get("use_directory_urls"),
                    )
                    self.i18n_files[language].append(i18n_ffile)
            else:
                # file is bundled by theme
                continue
        else:

            i18n_ffile = I18nFolderFile(
                fileobj,
                file_locale,
                all_languages=self.all_languages,
                default_language=self.default_language,
                docs_dir=config["docs_dir"],
                site_dir=config["site_dir"],
                use_directory_urls=config.get("use_directory_urls"),
            )
            self.i18n_files[file_locale].append(i18n_ffile)

            if file_locale == self.default_language:
                i18n_ffile = I18nFolderFile(
                    fileobj,
                    "",
                    all_languages=self.all_languages,
                    default_language=self.default_language,
                    docs_dir=config["docs_dir"],
                    site_dir=config["site_dir"],
                    use_directory_urls=config.get("use_directory_urls"),
                )
                main_files.append(i18n_ffile)

    # these comments are here to help me debug later if needed
    # print([{p.src_path: p.url} for p in main_files.documentation_pages()])
    # print([{p.src_path: p.url} for p in self.i18n_files["en"].documentation_pages()])
    # print([{p.src_path: p.url} for p in self.i18n_files["fr"].documentation_pages()])
    # print([{p.src_path: p.url} for p in main_files.static_pages()])
    # print([{p.src_path: p.url} for p in self.i18n_files["en"].static_pages()])
    # print([{p.src_path: p.url} for p in self.i18n_files["fr"].static_pages()])

    # populate pages alternates
    # main default version
    for page in main_files.documentation_pages():
        for language in self.all_languages:
            # do not list languages not being built as alternates
            if self.config["languages"].get(language, {}).get("build", False) is False:
                continue
            alternate = self.i18n_files[language].get_localized_page_from_url(
                page.url, language
            )
            if alternate:
                page.alternates[language] = alternate
            else:
                log.warning(
                    f"could not find '{language}' alternate for the default version of page '{page.src_path}'"
                )

    return main_files


def on_nav(self, nav, config, files):
    """ """
    # translate default nav, see #113
    if self._maybe_translate_titles(self.default_language, nav):
        log.info(f"Translated default navigation to {self.default_language}")

    # check if the navigation is manually configured, see #145
    manual_nav = config.get("nav") is not None

    for language, lang_config in self.config["languages"].items():
        # skip nav generation for languages that we do not build
        if lang_config["build"] is False:
            continue
        if self.i18n_configs[language]["nav"]:
            self._fix_config_navigation(language, self.i18n_files[language])

        self.i18n_navs[language] = get_navigation(
            self.i18n_files[language], self.i18n_configs[language]
        )
        if manual_nav is False:
            if self.i18n_navs[language].items[0].children is None:
                # the structure is weird, say it but do not crash hard, see #152
                log.warning(
                    f"The structure of folder '{config['docs_dir']}/{language}' "
                    "does not look right, expect navigation or url inconsistencies"
                )
            else:
                # the expected folder structure starts with a [Section(title='LANG')]
                # so we render our navigation using it as a root
                self.i18n_navs[language].items = (
                    self.i18n_navs[language].items[0].children
                )
            for item in self.i18n_navs[language]:
                if config["use_directory_urls"] is True:
                    expected_url = f"{language}/"
                else:
                    expected_url = f"{language}/index.html"
                if item.is_page and item.url.strip() == expected_url:
                    self.i18n_navs[language].homepage = item
                    break
            else:
                raise Exception(f"could not find homepage Page(url='{expected_url}')")

        # If awesome-pages is used, we want to use it to structurate our
        # localized navigations as well
        if "awesome-pages" in config["plugins"]:
            self.i18n_navs[language] = config["plugins"]["awesome-pages"].on_nav(
                self.i18n_navs[language],
                config=self.i18n_configs[language],
                files=self.i18n_files[language],
            )

        if self.config["nav_translations"].get(language, {}):
            if self._maybe_translate_titles(language, self.i18n_navs[language]):
                log.info(f"Translated navigation to {language}")

        if language == self.default_language:
            for section in nav.items:
                if section.title == self.default_language.capitalize():
                    nav.items = section.children
                    break
            else:
                if manual_nav is False:
                    raise Exception(
                        f"could not find default version Section(title='{self.default_language.capitalize()}')"
                    )
            for item in nav:
                if config["use_directory_urls"] is True:
                    expected_url = ""
                else:
                    expected_url = "index.html"
                if item.is_page and item.url == expected_url:
                    nav.homepage = item
                    break
            else:
                raise Exception(
                    f"could not find default homepage Page(url='{expected_url}')"
                )

    return nav



import logging
from collections import defaultdict
from copy import deepcopy
from pathlib import Path

from mkdocs import __version__ as mkdocs_version
from mkdocs import plugins
from mkdocs.commands.build import _build_page, _populate_page
from mkdocs.config.config_options import Choice, Type
from mkdocs.plugins import BasePlugin

import mkdocs_static_i18n.folder_structure as folder_structure
import mkdocs_static_i18n.suffix_structure as suffix_structure
from mkdocs_static_i18n import __file__ as installation_path
from mkdocs_static_i18n.structure import Locale

try:
    from mkdocs.localization import install_translations
except ImportError:
    install_translations = None

try:
    import pkg_resources

    material_dist = pkg_resources.get_distribution("mkdocs-material")
    material_version = material_dist.version
    material_languages = [
        lang.split(".html")[0]
        for lang in material_dist.resource_listdir("material/partials/languages")
    ]
except Exception:
    material_languages = []
    material_version = None

log = logging.getLogger("mkdocs.plugins." + __name__)

LUNR_LANGUAGES = [
    "ar",
    "da",
    "de",
    "en",
    "es",
    "fi",
    "fr",
    "hu",
    "it",
    "ja",
    "nl",
    "no",
    "pt",
    "ro",
    "ru",
    "sv",
    "th",
    "tr",
    "vi",
]
MKDOCS_THEMES = ["mkdocs", "readthedocs"]


class I18n(BasePlugin):

    config_scheme = (
        ("default_language_only", Type(bool, default=False, required=False)),
        ("default_language", Locale(str, required=True)),
        (
            "docs_structure",
            Choice(["folder", "suffix"], default="suffix", required=False),
        ),
        ("languages", Locale(dict, required=True)),
        ("material_alternate", Type(bool, default=True, required=False)),
        ("nav_translations", Type(dict, default={}, required=False)),
        ("search_reconfigure", Type(bool, default=True, required=False)),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.i18n_configs = {}
        self.i18n_files = defaultdict(list)
        self.i18n_navs = {}
        self.material_alternates = None

    @staticmethod
    def _is_url(value):
        return value.startswith("http://") or value.startswith("https://")

    def _dict_replace_value(self, directory, old, new):
        """
        Return a copy of the given dict with value replaced.
        """
        x = {}
        for k, v in directory.items():
            if isinstance(v, dict):
                v = self._dict_replace_value(v, old, new)
            elif isinstance(v, list):
                v = self._list_replace_value(v, old, new)
            elif isinstance(v, str) or isinstance(v, Path):
                if str(v) == str(old):
                    v = new
                if not self._is_url(v):
                    v = str(Path(v))
            x[k] = v
        return x

    def _list_replace_value(self, listing, old, new):
        """
        Return a copy of the given list with value replaced.
        """
        x = []
        for e in listing:
            if isinstance(e, list):
                e = self._list_replace_value(e, old, new)
            elif isinstance(e, dict):
                e = self._dict_replace_value(e, old, new)
            elif isinstance(e, str) or isinstance(e, Path):
                if str(e) == str(old):
                    e = new
                if not self._is_url(e):
                    e = str(Path(e))
            x.append(e)
        return x

    def _set_languages_options(self, config):
        """
        Configure languages options for the 'default' language
        """
        # Set the 'site_name' for all configured languages
        for language, lang_config in self.config["languages"].items():
            localized_site_name = lang_config["site_name"] or config["site_name"]
            self.config["languages"][language]["site_name"] = localized_site_name
        # the default_language options can be made available for the
        # 'default' / version of the website
        self.default_language_options = self.config["languages"].pop(
            "default",
            {
                "name": "default",
                "link": "./",
                "fixed_link": None,
                "build": True,
                "site_name": config["site_name"],
            },
        )
        if self.default_language_options["name"] == "default":
            default_language_config = self.config["languages"].get(
                self.default_language, {}
            )
            self.default_language_options["name"] = default_language_config.get(
                "name", self.default_language
            )
            self.default_language_options["site_name"] = default_language_config.get(
                "site_name", config["site_name"]
            )
            self.default_language_options["fixed_link"] = default_language_config.get(
                "fixed_link", None
            )
        # when the default language is listed on the languages
        # this means that the user wants a /default_language version
        # of his website
        if self.default_language not in self.config["languages"]:
            # no other language than default language set?
            if len(self.config["languages"]) == 0:
                build = True
            else:
                build = False
            self.config["languages"][self.default_language] = {
                "name": self.default_language_options["name"],
                "link": "./",
                "fixed_link": None,
                "build": build,
                "site_name": config["site_name"],
            }

    def on_config(self, config, **kwargs):
        """
        Enrich configuration with language specific knowledge.
        """
        self.default_language = self.config["default_language"]
        self._set_languages_options(config)
        # Make an order preserving list of all the configured languages
        self.all_languages = []
        for language in self.config["languages"]:
            if language not in self.all_languages:
                self.all_languages.append(language)
        if self.default_language not in self.all_languages:
            self.all_languages.insert(0, self.default_language)
        # Get our placement index in the plugins config list
        try:
            i18n_index = list(config["plugins"].keys()).index("i18n")
        except ValueError:
            i18n_index = -1
        # Make sure with-pdf is controlled by us, see #110
        # We will only control it for the main language, localized PDF are
        # generated on the 'on_post_build' method
        if "with-pdf" in config["plugins"]:
            with_pdf_index = list(config["plugins"].keys()).index("with-pdf")
            if with_pdf_index > i18n_index:
                config["plugins"]["with-pdf"].on_config(config)
            if mkdocs_version >= "1.4":
                # TODO: drop me after we start using mkdocs plugin prioritization
                log.warning(
                    "The 'with-pdf' plugin should be listed AFTER 'i18n' in the 'plugins' option"
                )
            else:
                for events in config["plugins"].events.values():
                    config["plugins"].move_to_end("with-pdf", last=False)
                    for idx, event in enumerate(list(events)):
                        try:
                            if str(event.__module__) == "mkdocs_with_pdf.plugin":
                                events.pop(idx)
                        except AttributeError:
                            # partials don't have a module
                            pass
        # Make sure awesome-pages is always called before us, see #65
        if "awesome-pages" in config["plugins"]:
            awesome_index = list(config["plugins"].keys()).index("awesome-pages")
            if awesome_index > i18n_index:
                if mkdocs_version >= "1.4":
                    # TODO: drop me after we start using mkdocs plugin prioritization
                    log.warning(
                        "The 'awesome-pages' plugin should be listed BEFORE 'i18n' in the 'plugins' option"
                    )
                else:
                    config["plugins"].move_to_end("awesome-pages", last=False)
                    for events in config["plugins"].events.values():
                        for idx, event in enumerate(list(events)):
                            try:
                                if (
                                    str(event.__module__)
                                    == "mkdocs_awesome_pages_plugin.plugin"
                                ):
                                    events.insert(0, events.pop(idx))
                            except AttributeError:
                                # partials don't have a module
                                pass
        # Make a localized copy of the config
        # The hooks are mutualized so we remove them from the config before (deep)copying
        # The plugins are mutualized so we remove them from the config before (deep)copying
        if "hooks" in config:
            hooks = config.pop("hooks")
        else:
            hooks = None
        plugins = config.pop("plugins")
        for language in self.all_languages:
            self.i18n_configs[language] = deepcopy(config)
            self.i18n_configs[language]["plugins"] = plugins
            if hooks:
                self.i18n_configs[language]["hooks"] = hooks
        config["plugins"] = plugins
        if hooks:
            config["hooks"] = hooks
        # Set theme locale to default language
        if self.default_language != "en":
            if config["theme"].name in MKDOCS_THEMES:
                if mkdocs_version >= "1.2":
                    config["theme"]["locale"] = self.default_language
                    log.info(
                        f"Setting the default 'theme.locale' option to '{self.default_language}'"
                    )
            elif config["theme"].name == "material":
                config["theme"].language = self.default_language
                log.info(
                    f"Setting the default 'theme.language' option to '{self.default_language}'"
                )
        # Skip language builds requested?
        if self.config["default_language_only"] is True:
            return config
        # Support for mkdocs-material>=7.1.0 language selector
        if self.config["material_alternate"] and len(self.all_languages) > 1:
            if material_version and material_version >= "7.1.0":
                if not config["extra"].get("alternate") or kwargs.get("force"):
                    config["extra"]["alternate"] = []
                    # Add index.html file name when used with
                    # use_directory_urls = True
                    link_suffix = ""
                    if config.get("use_directory_urls") is False:
                        link_suffix = "index.html"
                    # TODO: document
                    if self.default_language_options["build"] is True:
                        config["extra"]["alternate"].append(
                            {
                                "name": f"{self.default_language_options['name']}",
                                "link": f"{self.default_language_options['link']}{link_suffix}",
                                "fixed_link": self.default_language_options[
                                    "fixed_link"
                                ],
                                "lang": self.default_language,
                            }
                        )
                    for language, lang_config in self.config["languages"].items():
                        if lang_config["build"] is True:
                            if (
                                self.default_language_options["build"] is True
                                and lang_config["name"]
                                == self.default_language_options["name"]
                            ):
                                continue
                            config["extra"]["alternate"].append(
                                {
                                    "name": f"{lang_config['name']}",
                                    "link": f"{lang_config['link']}{link_suffix}",
                                    "fixed_link": lang_config["fixed_link"],
                                    "lang": language,
                                }
                            )
                elif "alternate" in config["extra"]:
                    for alternate in config["extra"]["alternate"]:
                        if not alternate.get("link", "").startswith("./"):
                            log.info(
                                "The 'extra.alternate' configuration contains a "
                                "'link' option that should starts with './' in "
                                f"{alternate}"
                            )

                if "navigation.instant" in config["theme"]._vars.get("features", []):
                    log.warning(
                        "mkdocs-material language switcher contextual link is not "
                        "compatible with theme.features = navigation.instant"
                    )
                else:
                    self.material_alternates = config["extra"].get("alternate")
        # Support for the search plugin lang
        if self.config["search_reconfigure"] and "search" in config["plugins"]:
            search_langs = config["plugins"]["search"].config["lang"] or []
            for language in self.all_languages:
                if language in LUNR_LANGUAGES:
                    if language not in search_langs:
                        search_langs.append(language)
                        log.info(
                            f"Adding '{language}' to the 'plugins.search.lang' option"
                        )
                else:
                    log.warning(
                        f"Language '{language}' is not supported by "
                        f"lunr.js, not setting it in the 'plugins.search.lang' option"
                    )
        # Report misconfigured nav_translations, see #66
        if self.config["nav_translations"]:
            for lang in self.config["languages"]:
                if lang in self.config["nav_translations"]:
                    break
            else:
                log.info(
                    "Ignoring 'nav_translations' option: expected a language key "
                    f"from {list(self.config['languages'].keys())}, got "
                    f"{list(self.config['nav_translations'].keys())}"
                )
                self.config["nav_translations"] = {}
        # Install a i18n aware version of sitemap.xml if not provided by the user
        if not Path(
            Path(config["theme"]._vars.get("custom_dir", ".")) / Path("sitemap.xml")
        ).exists():
            custom_i18n_sitemap_dir = Path(
                Path(installation_path).parent / Path("custom_i18n_sitemap")
            ).resolve()
            config["theme"].dirs.insert(0, str(custom_i18n_sitemap_dir))
        return config

    def on_files(self, files, config):
        """
        Construct the main + lang specific file tree which will be used to
        generate the navigation for the default site and per language.
        """
        if self.config["docs_structure"] == "suffix":
            return suffix_structure.on_files(self, files, config)
        else:
            return folder_structure.on_files(self, files, config)

    def _fix_config_navigation(self, language, files):
        """
        When a static navigation is set in mkdocs.yml a user will usually
        structurate its navigation using the main (default language)
        documentation markdown pages.

        This function localizes the given pages to their translated
        counterparts if available.
        """
        for i18n_page in files.documentation_pages():
            if Path(i18n_page.src_path).suffixes == [f".{language}", ".md"]:
                config_path_expects = [
                    i18n_page.non_i18n_src_path.with_suffix(".md"),
                    i18n_page.non_i18n_src_path.with_suffix(
                        f".{self.default_language}.md"
                    ),
                ]
                for config_path in config_path_expects:
                    self.i18n_configs[language]["nav"] = self._list_replace_value(
                        self.i18n_configs[language]["nav"],
                        config_path,
                        i18n_page.src_path,
                    )

    def _maybe_translate_titles(self, language, items):
        translated = False
        translated_nav = self.config["nav_translations"].get(language, {})
        if translated_nav:
            for item in items:
                if hasattr(item, "title") and item.title in translated_nav:
                    log.debug(
                        f"Translating {type(item).__name__} title '{item.title}' "
                        f"({self.default_language}) to "
                        f"'{translated_nav[item.title]}' ({language})"
                    )
                    item.title = translated_nav[item.title]
                    translated = True
                if hasattr(item, "children") and item.children:
                    translated = (
                        self._maybe_translate_titles(language, item.children)
                        or translated
                    )
        return translated

    def on_nav(self, nav, config, files):
        """
        Translate i18n aware navigation to honor the 'nav_translations' option.
        """
        if self.config["docs_structure"] == "suffix":
            nav = suffix_structure.on_nav(self, nav, config, files)
        else:
            nav = folder_structure.on_nav(self, nav, config, files)
        # Manually trigger with-pdf on_nav, see #110
        with_pdf_plugin = config["plugins"].get("with-pdf")
        if with_pdf_plugin:
            with_pdf_plugin.on_nav(nav, config, files)
        return nav

    def _fix_search_duplicates(self, search_plugin):
        """
        We want to avoid indexing the same pages twice if the default language
        has its own version built as well as the /language version too as this
        would pollute the search results.

        When this happens, we favor the default language location if its
        content is the same as its /language counterpart.
        """
        default_lang_entries = filter(
            lambda x: not x["location"].startswith(
                tuple(self.config["languages"].keys())
            ),
            search_plugin.search_index._entries,
        )
        target_lang_entries = list(
            filter(
                lambda x: x["location"].startswith(
                    tuple(self.config["languages"].keys())
                ),
                search_plugin.search_index._entries,
            )
        )
        for default_lang_entry in default_lang_entries:
            duplicated_entries = filter(
                lambda x: x["title"] == default_lang_entry["title"]
                and x["location"].endswith(x["location"])
                and x["text"] == default_lang_entry["text"],
                target_lang_entries,
            )
            for duplicated_entry in duplicated_entries:
                if duplicated_entry in search_plugin.search_index._entries:
                    log.debug(
                        f"removed duplicated search entry: {duplicated_entry['title']} "
                        f"{duplicated_entry['location']}"
                    )
                    search_plugin.search_index._entries.remove(duplicated_entry)

    @plugins.event_priority(-100)
    def on_env(self, env, config, files, **kwargs):
        """
        Copy the main env because it can have been altered by hooks.

        Since hooks are injected as plugins, we need to make sure to
        run last using event_priority.

        See #178.
        """
        for language in self.config["languages"].keys():
            self.i18n_configs[language]["env"] = env
        return env

    def on_page_markdown(self, markdown, page, config, files):
        """
        Use the 'page_markdown' event to translate page titles as well
        because they are used in the bottom navigation links for
        previous/next links titles.

        This event is triggered right after the page.read_source() has
        been executed so we get the definitive 'page.title' that mkdocs
        discovered and we just have to translate if it needed.
        """
        page.locale = page.file.dest_language or self.default_language
        if self.config["nav_translations"].get(page.locale, {}):
            self._maybe_translate_titles(page.locale, [page])
        return markdown

    def on_page_context(self, context, page, config, nav):
        """
        Make the language switcher contextual to the current page.

        This allows to switch language while staying on the same page.
        """
        # export some useful i18n related variables on page context, see #75
        context["i18n_config"] = self.config
        context["i18n_page_locale"] = page.locale
        context["i18n_page_file_locale"] = page.file.locale_suffix

        if self.material_alternates:
            alternates = deepcopy(self.material_alternates)
            page_url = page.url
            for language in self.all_languages:
                if page.url.startswith(f"{language}/"):
                    prefix_len = len(language) + 1
                    page_url = page.url[prefix_len:]
                    break

            for alternate in alternates:
                if config.get("use_directory_urls") is False:
                    alternate["link"] = alternate["link"].replace("/index.html", "", 1)
                fixed_link = alternate.get("fixed_link")
                if fixed_link:
                    alternate["link"] = fixed_link
                else:
                    if not alternate["link"].endswith("/"):
                        alternate["link"] += "/"
                    alternate["link"] += page_url
            config["extra"]["alternate"] = alternates

        # set the localized site_name if any
        if page.file.dest_language == "":
            # default
            localized_site_name = self.default_language_options["site_name"]
        else:
            localized_site_name = (
                self.config["languages"]
                .get(context["i18n_page_locale"], {})
                .get("site_name", config["site_name"])
            )
        config["site_name"] = localized_site_name

        return context

    def on_post_page(self, output, page, config):
        """
        Some plugins we control ourselves need this event.
        """
        # Manually trigger with-pdf on_nav, see #110
        with_pdf_plugin = config["plugins"].get("with-pdf")
        if with_pdf_plugin:
            with_pdf_plugin.on_post_page(output, page, config)
        return output

    def on_post_build(self, config):
        """
        Derived from mkdocs commands build function.

        We build every language on its own directory.
        """
        # user requested only the default version to be built
        if self.config["default_language_only"] is True:
            return

        dirty = False
        minify_plugin = config["plugins"].get("minify")
        search_plugin = config["plugins"].get("search")
        with_pdf_plugin = config["plugins"].get("with-pdf")
        if with_pdf_plugin:
            with_pdf_plugin.on_post_build(config)
            with_pdf_output_path = with_pdf_plugin.config["output_path"]
        for language, language_config in self.config["languages"].items():
            # Language build disabled by the user, skip
            if language_config["build"] is False:
                log.info(f"Skipping building {language} documentation")
                continue

            log.info(f"Building {language} documentation")

            config = self.i18n_configs[language]
            env = self.i18n_configs[language]["env"]
            files = self.i18n_files[language]
            nav = self.i18n_navs[language]

            # Support mkdocs-material theme language
            if config["theme"].name == "material":
                if language in material_languages:
                    config["theme"].language = language
                else:
                    log.warning(
                        f"Language {language} is not supported by "
                        f"mkdocs-material=={material_version}, not setting "
                        "the 'theme.language' option"
                    )

            # Support mkdocs-minify-plugin
            if minify_plugin:
                minify_plugin.on_pre_build(config=config)

            # Include theme specific files
            files.add_files_from_theme(env, config)

            # Include static files
            files.copy_static_files(dirty=dirty)

            for file in files.documentation_pages():
                _populate_page(file.page, config, files, dirty)

            for file in files.documentation_pages():
                _build_page(file.page, config, files, nav, env, dirty)

            if with_pdf_plugin:
                with_pdf_plugin.config[
                    "output_path"
                ] = f"{language}/{with_pdf_output_path}"
                with_pdf_plugin.on_config(config)
                with_pdf_plugin.on_nav(nav, config, files)
                with_pdf_plugin.on_post_build(config)

        # Update the search plugin index with language pages
        if search_plugin:
            self._fix_search_duplicates(search_plugin)
            search_plugin.on_post_build(config)
